//
//  NisusRangeIterator.h
//  NisusCore
//
//  Created by Martin on 2017.12.11.
//  Copyright © 2017 Nisus. All rights reserved.
//
//	This data structure provides an easy way to iterate a range via index.

#import <Foundation/Foundation.h>
#import <NisusCore/NisusRangeUtil.h>


typedef struct {
	NSUInteger nextIndex;
	NSUInteger boundIndex;
	BOOL isForwards;
} NisusRangeIterator;




// Constructs a new iterator that will visit the given range.
// ie: calling NisusRangeIterator_NextRange() will return the given range.
static NISUS_INLINE NisusRangeIterator NisusRangeIterator_Make( NSRange range, BOOL isForwards )
{
	NisusRangeIterator iter;
	iter.isForwards = isForwards;
	iter.nextIndex = isForwards ? range.location : NSMaxRange(range);
	iter.boundIndex = isForwards ? NSMaxRange(range) : range.location;
	return iter;
}

// Returns YES if the given iterator is exhausted and there is no remaining range.
static NISUS_INLINE BOOL NisusRangeIterator_IsEmpty( NisusRangeIterator const * const iter )
{
	if( iter->isForwards ) {
		return (iter->nextIndex >= iter->boundIndex);
	}
	else {
		return (iter->nextIndex <= iter->boundIndex);
	}
}

// Returns the range left in the iterator that hasn't been skipped yet.
// Returns a zero length range if already exhausted.
static NISUS_INLINE NSRange NisusRangeIterator_NextRange( NisusRangeIterator const * const iter )
{
	NSRange remRange;
	
	if( iter->isForwards ) {
		if( iter->nextIndex >= iter->boundIndex ) {
			remRange = NSMakeRange( iter->boundIndex, 0 );
		}
		else {
			remRange = NisusRange_AtWithBound( iter->nextIndex, iter->boundIndex );
		}
	}
	else {
		if( iter->nextIndex <= iter->boundIndex ) {
			remRange = NSMakeRange( iter->boundIndex, 0 );
		}
		else {
			remRange = NisusRange_AtWithBound( iter->boundIndex, iter->nextIndex );
		}
	}
	
	return remRange;
}

static NISUS_INLINE BOOL NisusRangeIterator_GetNextRange( NisusRangeIterator const * const iter, NSRange* nextRangePtr )
{
	NSRange nextRange = NisusRangeIterator_NextRange(iter);
	*nextRangePtr = nextRange;
	return (nextRange.length > 0);
}

// Returns the first index in the next range, or NSNotFound if the iterator is empty.
static NISUS_INLINE NSUInteger NisusRangeIterator_NextIndex( NisusRangeIterator const * const iter )
{
	if( iter->isForwards ) {
		if( iter->nextIndex >= iter->boundIndex ) {
			return NSNotFound;
		}
		else {
			return iter->nextIndex;
		}
	}
	else {
		if( iter->nextIndex <= iter->boundIndex ) {
			return NSNotFound;
		}
		else {
			return iter->nextIndex - 1;
		}
	}
}

// Advances the iterator so it skips the given range, ie: moves to NSMaxRange(skipRange)
static NISUS_INLINE void NisusRangeIterator_SkipRange( NisusRangeIterator* iter, NSRange skipRange )
{
	if( iter->isForwards ) {
		iter->nextIndex = NSMaxRange(skipRange);
	}
	else {
		iter->nextIndex = skipRange.location;
	}
}

// Sets the iterator so the next range it returns includes the given location.
// If the given location is outside the original range then this sets the iterator to empty.
static NISUS_INLINE void NisusRangeIterator_SetNextIndex( NisusRangeIterator* iter, NSUInteger setIndex )
{
	if( iter->isForwards ) {
		iter->nextIndex = setIndex;
	}
	else {
		iter->nextIndex = setIndex + 1;
	}
}

static NISUS_INLINE void _NisusRangeIterator_AdjustIndexForEdit( NSUInteger* charIndexPtr, BOOL isForwards, NSRange editedRange, NSInteger delta, BOOL* hadJumpPtr )
{
	NSUInteger charIndex = *charIndexPtr;
	
	NSRange origEditRange = NSMakeRange( editedRange.location, editedRange.length - delta );
	if( charIndex <= origEditRange.location ) {
		// index is unaffected; the edit occurred after it
	}
	else if( charIndex < NSMaxRange(origEditRange) ) {
		// index occurred within the edited range, so we don't know how it was affected exactly. Just jump to the end/start of the valid edited range
		*charIndexPtr = isForwards ? NSMaxRange(editedRange) : editedRange.location;
		*hadJumpPtr = YES;
	}
	else {
		// index occurred after the edited range, so shift it
		NISUS_ASSERT( (delta >= 0) || (ABS(delta) <= charIndex), "range iterator" );
		*charIndexPtr += delta;
	}
}

// This function adjusts the iterator to account for an edit within its original range. The iterator will always remain valid.
// However, if the edited range encompasses the current index or boundary condition, some indexes may be skipped.
// Returns YES if adjusted perfectly (no possible skipping), or returns NO if a skip may have occurred.
static NISUS_INLINE BOOL NisusRangeIterator_AdjustForEdit( NisusRangeIterator* iter, NSRange editedRange, NSInteger delta )
{
	NISUS_ASSERT( (delta <= 0) || (delta <= editedRange.length), "range iterator" );
	
	BOOL hadJump = NO;
	if( 0 != delta ) {
		BOOL isForwards = iter->isForwards;
		_NisusRangeIterator_AdjustIndexForEdit( &iter->nextIndex, isForwards, editedRange, delta, &hadJump );
		_NisusRangeIterator_AdjustIndexForEdit( &iter->boundIndex, isForwards, editedRange, delta, &hadJump );
	}
	return ! hadJump;
}

// This function visits all remaining effective ranges in the given iterator.
static NISUS_INLINE void NisusRangeIterator_VisitEffectiveRanges( NisusRangeIterator* iter, NSRange(^effRangeCalc)(NSUInteger testCharIndex, BOOL* stop) )
{
	while( YES ) {
		NSUInteger charIndex = NisusRangeIterator_NextIndex(iter);
		if( NSNotFound == charIndex ) {
			break;
		}
		
		BOOL stop = NO;
		NSRange effRange = effRangeCalc( charIndex, &stop );
		if( stop ) {
			break;
		}
		
		NisusRangeIterator_SkipRange( iter, effRange );
	}
}

// Convenience function to scan a range using arbitrary jumps. The block is repeatedly called with the remainder of the original range.
// The block should return the first index of the NEXT range to visit, eg: NSMaxRange of the effective range just visited and processed.
// It is considered an error if the visitor fails to make progress (advances zero indexes) or goes backwards. That will trigger an abort.
static NISUS_INLINE void NisusRange_VisitForwards( NSRange visitRange, NSUInteger(NS_NOESCAPE^visitor)(NSRange remRange, BOOL* stop) )
{
	BOOL stop = NO;
	NSUInteger scanAt = visitRange.location;
	NSUInteger scanLim = NSMaxRange(visitRange);
	
	while( scanAt < scanLim ) {
		NSRange remRange = NisusRange_AtWithBound( scanAt, scanLim );
		
		NSUInteger nextIndex = visitor( remRange, &stop );
		if( stop ) {
			break;
		}
		NISUS_ASSERT_SAFE( nextIndex > scanAt, "range visitor", break );
		scanAt = nextIndex;
	}
}




